class AbsAny:
    """
    Any class defined in this project that is made to be a hdict value or should not be used as such

    This intended to ease checking disallowed types being mistakengly used as user values.
    """
